/****************************************************************************
 *																			*
 * File    : Worker.c														*
 *																			*
 * Purpose : Parse a COM Type Library and build type definitions.			*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

/*
 * This is a sample. Not all types are supported, and some of the algorithms
 * are very simplistic. The good news is that you got the source - so all
 * this can be fixed! ;-)
 */

#define WIN32_LEAN_AND_MEAN
#define UNICODE   /* for Windows API */
#define _UNICODE  /* for C runtime */
#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <assert.h>
#include "TypeLib.h"

#define COBJMACROS
#include <ole2.h>

/* formats for _*tprintf() */
#ifdef _UNICODE
#define _TSFMT	L"%ls"
#define _TCFMT	L"%lc"
#else
#define _TSFMT	"%s"
#define _TCFMT	"%c"
#endif /* !_UNICODE */

#define MAXLINE  1024
#define BUFSTEP  (MAXLINE * 100)

/* Output buffer */
typedef struct tagOUTPUT {
	PTSTR pchBuf;			/* Pointer to buffer, or NULL */
	DWORD cchBuf;			/* Current number of chars */
	DWORD cchMaxBuf;		/* Maximum number of chars */
} OUTPUT, *POUTPUT;

/* Linked list for tag names */
typedef struct tagNAME {
	struct tagNAME *pNext;	/* Pointer to next node, or NULL */
	WCHAR *pwszName;		/* Tag name (Unicode, since COM loves that) */
} NAME, *PNAME;

/* Head of linked list */
static PNAME g_pNameHead = NULL;
static OUTPUT Fwd = {0};

/* Static function prototypes */
static void EnumTypeLib(POUTPUT, LPTYPELIB);
static void DumpTypeInfo(POUTPUT, LPTYPEINFO, BOOL);
static void DumpAliasType(POUTPUT, BSTR, LPTYPEATTR, LPTYPEINFO);
static void DumpClassType(POUTPUT, BSTR, LPTYPEATTR);
static void DumpEnumType(POUTPUT, BSTR, BSTR, LPTYPEATTR, LPTYPEINFO);
static void DumpRecordType(POUTPUT, BSTR, BSTR, LPTYPEATTR, LPTYPEINFO);
static void DumpInterfaceType(POUTPUT, BSTR, LPTYPEATTR, LPTYPEINFO, BOOL);
static void GetTypeName(PTSTR, int, TYPEDESC, BSTR, LPTYPEINFO);
static void StrCat(PTSTR, int, PCTSTR, ...);
static void BufCat(POUTPUT, PCTSTR, ...);
static PNAME AllocName(const WCHAR *);
static PNAME LookupName(const WCHAR *);

/****************************************************************************
 *																			*
 * Function: DumpTypeLib													*
 *																			*
 * Purpose : Dump a type library to a buffer, and return it.				*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

PTSTR DumpTypeLib(PCTSTR pszFilename)
{
	OUTPUT Out = {0};
	LPTYPELIB pITypeLib;
	DECLT2W

	memset(&Fwd, 0, sizeof(Fwd));

	__try
	{
		/* Load the given type library */
		HRESULT hr = LoadTypeLibEx(T2W(pszFilename), REGKIND_NONE, &pITypeLib);
		if (hr == S_OK)
		{
			TCHAR szName[MAX_PATH];
			PCTSTR psz, pszQ;
			DWORD cchNow;

			/* Find beginning of name */
			for (psz = pszFilename + _tcslen(pszFilename);
				 psz > pszFilename && psz[-1] != _T('\\') && psz[-1] != _T('/') && psz[-1] != _T(':');
				 psz--)
				;

			/* Find end of name */
			for (pszQ = psz; *pszQ != _T('\0') && *pszQ != _T('.'); pszQ++)
				;

			/* Extract name */
			lstrcpyn(szName, psz, pszQ - psz + 1);
			CharUpper(szName);

			BufCat(&Out, _T("#ifndef COM_NO_WINDOWS_H\n"));
			BufCat(&Out, _T("#include <windows.h>\n"));
			BufCat(&Out, _T("#include <ole2.h>\n"));
			BufCat(&Out, _T("#endif\n\n"));

			BufCat(&Out, _T("#ifndef _%ls_H\n"), szName);
			BufCat(&Out, _T("#define _%ls_H\n"), szName);

			/* Remember buffer offset */
			cchNow = Out.cchBuf;

			/* Enumerate types in the type library */
			EnumTypeLib(&Out, pITypeLib);

			BufCat(&Out, _T("\n#endif /* _") _TSFMT _T("_H */\n"), szName);

			/* If we have interfaces, merge the forward declarations into the main buffer */
			if (Fwd.cchBuf != 0)
			{
				PTSTR pch = malloc((Fwd.cchBuf + Out.cchBuf + 1) * sizeof(TCHAR));
				if (pch != NULL)
				{
					_tmemcpy(pch, Out.pchBuf, cchNow);
					_tmemcpy(pch + cchNow, Fwd.pchBuf, Fwd.cchBuf);
					_tmemcpy(pch + cchNow + Fwd.cchBuf, Out.pchBuf + cchNow, Out.cchBuf - cchNow);
					pch[Fwd.cchBuf + Out.cchBuf] = L'\0';
					free(Fwd.pchBuf); Fwd.pchBuf = NULL;
					free(Out.pchBuf); Out.pchBuf = pch;
				}
			}

			ITypeLib_Release(pITypeLib);

			/* Free linked list of tag names */
			while (g_pNameHead != NULL)
			{
				PNAME pName = g_pNameHead;
				g_pNameHead = g_pNameHead->pNext;
				free(pName);
			}
		}
	}
	__except (EXCEPTION_EXECUTE_HANDLER)
	{
		free(Fwd.pchBuf);
		free(Out.pchBuf);
		Fwd.pchBuf = Out.pchBuf = NULL;
	}

	/* Return the buffer */
	return Out.pchBuf;
}

/****************************************************************************
 *																			*
 * Function: EnumTypeLib													*
 *																			*
 * Purpose : Walk through a type library and dump it's types.				*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void EnumTypeLib(POUTPUT pOut, LPTYPELIB pITypeLib)
{
	UINT cTypes = ITypeLib_GetTypeInfoCount(pITypeLib);
	for (UINT i = 0; i < cTypes; i++)
	{
		LPTYPEINFO pITypeInfo;

		/* Dump this type, please */
		if (ITypeLib_GetTypeInfo(pITypeLib, i, &pITypeInfo) == S_OK)
		{
			DumpTypeInfo(pOut, pITypeInfo, FALSE);
			ITypeInfo_Release(pITypeInfo);
		}
	}
}

/****************************************************************************
 *																			*
 * Function: DumpTypeInfo													*
 *																			*
 * Purpose : Dump information about a specific type.						*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void DumpTypeInfo(POUTPUT pOut, LPTYPEINFO pITypeInfo, BOOL fDispatch)
{
	BSTR pwszTypeName, pwszComment;

	if (ITypeInfo_GetDocumentation(pITypeInfo, MEMBERID_NIL, &pwszTypeName, &pwszComment, 0, 0) == S_OK)
	{
		TYPEATTR *pTypeAttr;

		if (ITypeInfo_GetTypeAttr(pITypeInfo, &pTypeAttr) == S_OK)
		{
			/*
			 * Handle types we care about...
			 */
			switch (pTypeAttr->typekind)
			{
				static TYPEKIND TKindPrev = 0;

				case TKIND_COCLASS:
					if (pwszComment != NULL && *pwszComment != _T('\0'))
						BufCat(pOut, _T("\n/* %ls */\n"), pwszComment);
					else if (TKindPrev != TKIND_COCLASS)
						BufCat(pOut, _T("\n"));
					DumpClassType(pOut, pwszTypeName, pTypeAttr);
					TKindPrev = TKIND_COCLASS;
					break;

				case TKIND_ENUM:
					if (LookupName(pwszTypeName)) break;
					if (pwszComment != NULL && *pwszComment != _T('\0'))
						BufCat(pOut, _T("\n/* %ls */\n"), pwszComment);
					else
						BufCat(pOut, _T("\n"));
					DumpEnumType(pOut, pwszTypeName, pwszTypeName, pTypeAttr, pITypeInfo);
					break;

				case TKIND_RECORD:
				case TKIND_UNION:
					if (LookupName(pwszTypeName)) break;
					if (pwszComment != NULL && *pwszComment != _T('\0'))
						BufCat(pOut, _T("\n/* %ls */\n"), pwszComment);
					else
						BufCat(pOut, _T("\n"));
					DumpRecordType(pOut, pwszTypeName, pwszTypeName, pTypeAttr, pITypeInfo);
					break;

				case TKIND_ALIAS:
					if (pwszComment != NULL && *pwszComment != _T('\0'))
						BufCat(pOut, _T("\n/* %ls */\n"), pwszComment);
					else
						BufCat(pOut, _T("\n"));
					DumpAliasType(pOut, pwszTypeName, pTypeAttr, pITypeInfo);
					TKindPrev = TKIND_ALIAS;
					break;

				case TKIND_DISPATCH:
					if (pTypeAttr->wTypeFlags & TYPEFLAG_FDUAL)
					{
						HREFTYPE hreftype;
						ITypeInfo_GetRefTypeOfImplType(pITypeInfo, -1, &hreftype);
						ITypeInfo_GetRefTypeInfo(pITypeInfo, hreftype, &pITypeInfo);
						DumpTypeInfo(pOut, pITypeInfo, TRUE);
					}
					break;

				case TKIND_INTERFACE:
					BufCat(pOut, _T("\n"));
					BufCat(pOut, _T("#ifndef __%ls_INTERFACE_DEFINED__\n"), pwszTypeName);
					BufCat(pOut, _T("#define __%ls_INTERFACE_DEFINED__\n"), pwszTypeName);
					if (pwszComment != NULL && *pwszComment != _T('\0'))
						BufCat(pOut, _T("\n/* %ls */\n"), pwszComment);
					DumpInterfaceType(pOut, pwszTypeName, pTypeAttr, pITypeInfo, fDispatch);
					BufCat(pOut, _T("#endif /* __%ls_INTERFACE_DEFINED__ */\n"), pwszTypeName);
					break;
			}

			ITypeInfo_ReleaseTypeAttr(pITypeInfo, pTypeAttr);
		}

		SysFreeString(pwszTypeName);
		SysFreeString(pwszComment);
	}
}

/****************************************************************************
 *																			*
 * Function: DumpAliasType													*
 *																			*
 * Purpose : Dump information for TKIND_ALIAS.								*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void DumpAliasType(POUTPUT pOut, BSTR pwszTypeName, LPTYPEATTR pTypeAttr, LPTYPEINFO pITypeInfo)
{
	BOOL fDone = FALSE;

	if ((pTypeAttr->tdescAlias.vt & VT_TYPEMASK) == VT_USERDEFINED)
	{
		LPTYPEINFO pITypeInfo2;

		if (ITypeInfo_GetRefTypeInfo(pITypeInfo, pTypeAttr->tdescAlias.hreftype, &pITypeInfo2) == S_OK)
		{
			TYPEATTR *pTypeAttr2;

			if (ITypeInfo_GetTypeAttr(pITypeInfo2, &pTypeAttr2) == S_OK)
			{
				if (pTypeAttr2->typekind == TKIND_RECORD ||
					pTypeAttr2->typekind == TKIND_UNION ||
					pTypeAttr2->typekind == TKIND_ENUM)
				{
					BSTR pwszTypeName2;

					if (ITypeInfo_GetDocumentation(pITypeInfo2, MEMBERID_NIL, &pwszTypeName2, 0, 0, 0) == S_OK)
					{
						if (pTypeAttr2->typekind == TKIND_ENUM)
							DumpEnumType(pOut, pwszTypeName2, pwszTypeName, pTypeAttr2, pITypeInfo2);
						else
							DumpRecordType(pOut, pwszTypeName2, pwszTypeName, pTypeAttr2, pITypeInfo2);

						SysFreeString(pwszTypeName2);
					}

					fDone = TRUE;
				}

				ITypeInfo_ReleaseTypeAttr(pITypeInfo2, pTypeAttr2);
			}

			ITypeInfo_Release(pITypeInfo2);
		}
	}

	if (!fDone)
	{
		TCHAR szType[256] = _T("");

		GetTypeName(szType, NELEMS(szType), pTypeAttr->tdescAlias, NULL, pITypeInfo);
		BufCat(pOut, _T("typedef ") _TSFMT _T(" %ls;  /* ALIAS */\n"), szType, pwszTypeName);
	}
}

/****************************************************************************
 *																			*
 * Function: DumpClassType													*
 *																			*
 * Purpose : Dump information for TKIND_COCLASS.							*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void DumpClassType(POUTPUT pOut, BSTR pwszTypeName, LPTYPEATTR pTypeAttr)
{
	BufCat(pOut, _T("DEFINE_GUID(CLSID_%ls,0x%08X,0x%04X,0x%04X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X);\n"),
		pwszTypeName,
		pTypeAttr->guid.Data1,
		pTypeAttr->guid.Data2,
		pTypeAttr->guid.Data3,
		pTypeAttr->guid.Data4[0],
		pTypeAttr->guid.Data4[1],
		pTypeAttr->guid.Data4[2],
		pTypeAttr->guid.Data4[3],
		pTypeAttr->guid.Data4[4],
		pTypeAttr->guid.Data4[5],
		pTypeAttr->guid.Data4[6],
		pTypeAttr->guid.Data4[7]);
}

/****************************************************************************
 *																			*
 * Function: DumpEnumType													*
 *																			*
 * Purpose : Dump information for TKIND_ENUM.								*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void DumpEnumType(POUTPUT pOut, BSTR pwszTagName, BSTR pwszTypeName, LPTYPEATTR pTypeAttr, LPTYPEINFO pITypeInfo)
{
	AllocName(pwszTagName);

	/* Check, just in case */
	if (pTypeAttr->cVars != 0)
	{
		BufCat(pOut, _T("typedef enum %ls {\n"), pwszTagName);

		/* Walk through all elements */
		for (int i = 0; i < pTypeAttr->cVars; i++)
		{
			VARDESC *pVarDesc;

			/* Get general information for a single element */
			if (ITypeInfo_GetVarDesc(pITypeInfo, i, &pVarDesc) == S_OK)
			{
				BSTR pwszVarName, pwszComment;

				/* Get name and description for a single element */
				if (ITypeInfo_GetDocumentation(pITypeInfo, pVarDesc->memid, &pwszVarName, &pwszComment, 0, 0) == S_OK)
				{
					if (pVarDesc->varkind == VAR_CONST && V_VT(pVarDesc->lpvarValue) == VT_I4)
						BufCat(pOut, _T("\t%ls = %d,"), pwszVarName, V_I4(pVarDesc->lpvarValue));
					else
						BufCat(pOut, _T("\t%ls,"), pwszVarName);

					if (pwszComment != NULL && *pwszComment != _T('\0'))
						BufCat(pOut, _T("  /* %ls */\n"), pwszComment);
					else
						BufCat(pOut, _T("\n"));

					SysFreeString(pwszVarName);
					SysFreeString(pwszComment);
				}

				ITypeInfo_ReleaseVarDesc(pITypeInfo, pVarDesc);
			}
		}

		BufCat(pOut, _T("} %ls;\n"), pwszTypeName);
	}
}

/****************************************************************************
 *																			*
 * Function: DumpRecordType 												*
 *																			*
 * Purpose : Dump information for TKIND_RECORD or TKIND_UNION.				*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void DumpRecordType(POUTPUT pOut, BSTR pwszTagName, BSTR pwszTypeName, LPTYPEATTR pTypeAttr, LPTYPEINFO pITypeInfo)
{
	AllocName(pwszTagName);

	/* Check, just in case */
	if (pTypeAttr->cVars != 0)
	{
		BufCat(pOut, _T("typedef ") _TSFMT _T(" %ls {\n"), pTypeAttr->typekind == TKIND_UNION ? _T("union") : _T("struct"), pwszTagName);

		/* Walk through all members */
		for (int i = 0; i < pTypeAttr->cVars; i++)
		{
			VARDESC *pVarDesc;

			/* Get general information for a single member */
			if (ITypeInfo_GetVarDesc(pITypeInfo, i, &pVarDesc) == S_OK)
			{
				BSTR pwszVarName, pwszComment;

				/* Get name and description for a single member */
				if (ITypeInfo_GetDocumentation(pITypeInfo, pVarDesc->memid, &pwszVarName, &pwszComment, 0, 0) == S_OK)
				{
					TCHAR szType[256] = _T("");

					GetTypeName(szType, NELEMS(szType), pVarDesc->elemdescVar.tdesc, pwszVarName, pITypeInfo);
					BufCat(pOut, _T("\t") _TSFMT _T(";"), szType);

					if (pwszComment != NULL && *pwszComment != _T('\0'))
						BufCat(pOut, _T("  /* %ls */\n"), pwszComment);
					else
						BufCat(pOut, _T("\n"));

					SysFreeString(pwszVarName);
					SysFreeString(pwszComment);
				}

				ITypeInfo_ReleaseVarDesc(pITypeInfo, pVarDesc);
			}
		}

		BufCat(pOut, _T("} %ls;\n"), pwszTypeName);
	}
}

/****************************************************************************
 *																			*
 * Function: DumpInterfaceType												*
 *																			*
 * Purpose : Dump information for TKIND_INTERFACE.							*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void DumpInterfaceType(POUTPUT pOut, BSTR pwszTypeName, LPTYPEATTR pTypeAttr, LPTYPEINFO pITypeInfo, BOOL fDispatch)
{
	BufCat(pOut, _T("DEFINE_GUID(IID_%ls,0x%08X,0x%04X,0x%04X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X,0x%02X);\n"),
		pwszTypeName,
		pTypeAttr->guid.Data1,
		pTypeAttr->guid.Data2,
		pTypeAttr->guid.Data3,
		pTypeAttr->guid.Data4[0],
		pTypeAttr->guid.Data4[1],
		pTypeAttr->guid.Data4[2],
		pTypeAttr->guid.Data4[3],
		pTypeAttr->guid.Data4[4],
		pTypeAttr->guid.Data4[5],
		pTypeAttr->guid.Data4[6],
		pTypeAttr->guid.Data4[7]);

	// ? check for base interface (inheritance)
	BufCat(pOut, _T("\n#undef INTERFACE\n"));
	BufCat(pOut, _T("#define INTERFACE	%ls\n"), pwszTypeName);
	BufCat(pOut, _T("DECLARE_INTERFACE(%ls) {\n"), pwszTypeName);

	if (Fwd.cchBuf == 0) BufCat(&Fwd, _T("\n/* Forward declarations */\n"));
	BufCat(&Fwd, _T("typedef interface %ls %ls;\n"), pwszTypeName, pwszTypeName);

	BufCat(pOut, _T("\t// IUnknown methods\n"));
	BufCat(pOut, _T("\tSTDMETHOD(QueryInterface)(IDispatch*,REFIID,void**);\n"));
	BufCat(pOut, _T("\tSTDMETHOD_(ULONG,AddRef)(IDispatch*);\n"));
	BufCat(pOut, _T("\tSTDMETHOD_(ULONG,Release)(IDispatch*);\n"));

	if (fDispatch)
	{
		BufCat(pOut, _T("\t// IDispatch methods\n"));
		BufCat(pOut, _T("\tSTDMETHOD(GetTypeInfoCount)(IDispatch*,UINT*);\n"));
		BufCat(pOut, _T("\tSTDMETHOD(GetTypeInfo)(IDispatch*,UINT,LCID,ITypeInfo**);\n"));
		BufCat(pOut, _T("\tSTDMETHOD(GetIDsOfNames)(IDispatch*,REFIID,LPOLESTR*,UINT,LCID,DISPID*);\n"));
		BufCat(pOut, _T("\tSTDMETHOD(Invoke)(IDispatch*,DISPID,REFIID,LCID,WORD,DISPPARAMS*,VARIANT*,EXCEPINFO*,UINT*);\n"));
	}
	BufCat(pOut, _T("\t// %ls methods\n"), pwszTypeName);
	for (int i = 0; i < pTypeAttr->cFuncs; i++)
	{
		FUNCDESC *pFuncDesc;

		if (ITypeInfo_GetFuncDesc(pITypeInfo, i, &pFuncDesc) == S_OK)
		{
			BSTR pwszFuncName = 0;

			if (ITypeInfo_GetDocumentation(pITypeInfo, pFuncDesc->memid, &pwszFuncName, 0, 0, 0) == S_OK)
			{
				TCHAR szFuncName[256];
				TCHAR szType[256];

				_sntprintf(szFuncName, NELEMS(szFuncName), _TSFMT _T("%ls"), 
					pFuncDesc->invkind == INVOKE_PROPERTYGET ? _T("get_") :
					pFuncDesc->invkind == INVOKE_PROPERTYPUT ? _T("put_") :
					pFuncDesc->invkind == INVOKE_PROPERTYPUTREF ? _T("putref_") : _T(""),
					pwszFuncName);

				/* Get description of return type */
				*szType = _T('\0');
				GetTypeName(szType, NELEMS(szType), pFuncDesc->elemdescFunc.tdesc, NULL, pITypeInfo);

				if (_tcscmp(szType, _T("HRESULT")) == 0)
					BufCat(pOut, _T("\tSTDMETHOD(") _TSFMT _T(")"), szFuncName);
				else
					BufCat(pOut, _T("\tSTDMETHOD_(") _TSFMT _T(",") _TSFMT _T(")"), szType, szFuncName);

				if (pFuncDesc->cParams == 0)
					BufCat(pOut, _T("(THIS"));
				else
					BufCat(pOut, _T("(THIS_ "));

				for (int k = 0; k < pFuncDesc->cParams; k++)
				{
					*szType = _T('\0');
					GetTypeName(szType, NELEMS(szType), pFuncDesc->lprgelemdescParam[k].tdesc, NULL, pITypeInfo);
					BufCat(pOut, _TSFMT, szType);
					if (k < pFuncDesc->cParams-1)
						BufCat(pOut, _T(","));
				}

				BufCat(pOut, _T(");\n"));

				SysFreeString(pwszFuncName);
			}

			ITypeInfo_ReleaseFuncDesc(pITypeInfo, pFuncDesc);
		}
	}

	BufCat(pOut, _T("};\n\n"));
	BufCat(pOut, _T("#ifdef COBJMACROS\n"));

	BufCat(pOut, _T("#define %ls_QueryInterface(This,_1,_2)  (This)->lpVtbl->QueryInterface(This,_1,_2)\n"), pwszTypeName);
	BufCat(pOut, _T("#define %ls_AddRef(This)  (This)->lpVtbl->AddRef(This)\n"), pwszTypeName);
	BufCat(pOut, _T("#define %ls_Release(This)	(This)->lpVtbl->Release(This)\n"), pwszTypeName);

	if (fDispatch)
	{
		BufCat(pOut, _T("#define %ls_GetTypeInfoCount(This,_1)	(This)->lpVtbl->GetTypeInfoCount(This,_1)\n"), pwszTypeName);
		BufCat(pOut, _T("#define %ls_GetTypeInfo(This,_1,_2,_3)  (This)->lpVtbl->GetTypeInfo(This,_1,_2,_3)\n"), pwszTypeName);
		BufCat(pOut, _T("#define %ls_GetIDsOfNames(This,_1,_2,_3,_4,_5)  (This)->lpVtbl->GetIDsOfNames(This,_1,_2,_3,_4,_5)\n"), pwszTypeName);
		BufCat(pOut, _T("#define %ls_Invoke(This,_1,_2,_3,_4,_5,_6,_7,_8)  (This)->lpVtbl->Invoke(This,_1,_2,_3,_4,_5,_6,_7,_8)\n"), pwszTypeName);
	}

	for (int i = 0; i < pTypeAttr->cFuncs; i++)
	{
		FUNCDESC *pFuncDesc;

		if (ITypeInfo_GetFuncDesc(pITypeInfo, i, &pFuncDesc) == S_OK)
		{
			BSTR pwszFuncName = 0;

			if (ITypeInfo_GetDocumentation(pITypeInfo, pFuncDesc->memid, &pwszFuncName, 0, 0, 0) == S_OK)
			{
				TCHAR szFuncName[256];

				_sntprintf(szFuncName, NELEMS(szFuncName), _TSFMT _T("%ls"),
					pFuncDesc->invkind == INVOKE_PROPERTYGET ? _T("get_") :
					pFuncDesc->invkind == INVOKE_PROPERTYPUT ? _T("put_") :
					pFuncDesc->invkind == INVOKE_PROPERTYPUTREF ? _T("putref_") : _T(""),
					pwszFuncName);

				BufCat(pOut, _T("#define %ls_") _TSFMT _T("(This"), pwszTypeName, szFuncName);

				for (int k = 0; k < pFuncDesc->cParams; k++)
					BufCat(pOut, _T(",_%d"), k+1);

				BufCat(pOut, _T(")	(This)->lpVtbl->") _TSFMT _T("(This"), szFuncName);
				for (int k = 0; k < pFuncDesc->cParams; k++)
					BufCat(pOut, _T(",_%d"), k+1);

				BufCat(pOut, _T(")\n"));

				SysFreeString(pwszFuncName);
			}

			ITypeInfo_ReleaseFuncDesc(pITypeInfo, pFuncDesc);
		}
	}

	BufCat(pOut, _T("#endif /* COBJMACROS */\n\n"));
}

/****************************************************************************
 *																			*
 * Function: GetTypeName													*
 *																			*
 * Purpose : Get a C type from the given type description.					*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void GetTypeName(PTSTR pszBuf, int cchMaxBuf, TYPEDESC tdesc, BSTR pwszName, LPTYPEINFO pITypeInfo)
{
	if (tdesc.vt & VT_ARRAY)
		StrCat(pszBuf, cchMaxBuf, _T("SAFEARRAY("));

	switch (tdesc.vt & VT_TYPEMASK)
	{
		case VT_I2: StrCat(pszBuf, cchMaxBuf, _T("SHORT")); break;
		case VT_I4: StrCat(pszBuf, cchMaxBuf, _T("LONG")); break;
		case VT_R4: StrCat(pszBuf, cchMaxBuf, _T("float")); break;
		case VT_R8: StrCat(pszBuf, cchMaxBuf, _T("double")); break;
		case VT_CY: StrCat(pszBuf, cchMaxBuf, _T("CY")); break;
		case VT_DATE: StrCat(pszBuf, cchMaxBuf, _T("DATE")); break;
		case VT_BSTR: StrCat(pszBuf, cchMaxBuf, _T("BSTR")); break;
		case VT_DISPATCH: StrCat(pszBuf, cchMaxBuf, _T("IDispatch*")); break;
		case VT_ERROR: StrCat(pszBuf, cchMaxBuf, _T("SCODE")); break;
		case VT_BOOL: StrCat(pszBuf, cchMaxBuf, _T("VARIANT_BOOL")); break;
		case VT_VARIANT: StrCat(pszBuf, cchMaxBuf, _T("VARIANT")); break;
		case VT_UNKNOWN: StrCat(pszBuf, cchMaxBuf, _T("IUnknown*")); break;
		case VT_DECIMAL: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {decimal} */")); break;
		case VT_I1: StrCat(pszBuf, cchMaxBuf, _T("CHAR")); break;
		case VT_UI1: StrCat(pszBuf, cchMaxBuf, _T("UCHAR")); break;
		case VT_UI2: StrCat(pszBuf, cchMaxBuf, _T("USHORT")); break;
		case VT_UI4: StrCat(pszBuf, cchMaxBuf, _T("ULONG")); break;
		case VT_I8: StrCat(pszBuf, cchMaxBuf, _T("LONGLONG")); break;
		case VT_UI8: StrCat(pszBuf, cchMaxBuf, _T("ULONGLONG")); break;
		case VT_INT: StrCat(pszBuf, cchMaxBuf, _T("int")); break;
		case VT_UINT: StrCat(pszBuf, cchMaxBuf, _T("UINT")); break;
		case VT_VOID: StrCat(pszBuf, cchMaxBuf, _T("void")); break;
		case VT_HRESULT: StrCat(pszBuf, cchMaxBuf, _T("HRESULT")); break;
		case VT_SAFEARRAY: StrCat(pszBuf, cchMaxBuf, _T("SAFEARRAY*")); break;
		case VT_LPSTR: StrCat(pszBuf, cchMaxBuf, _T("char*")); break;
		case VT_LPWSTR: StrCat(pszBuf, cchMaxBuf, _T("WCHAR*")); break;
		case VT_RECORD: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {record} */")); break;
		case VT_FILETIME: StrCat(pszBuf, cchMaxBuf, _T("FILETIME")); break;
		case VT_BLOB: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {blob} */")); break;
		case VT_STREAM: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {stream} */")); break;
		case VT_STORAGE: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {storage} */")); break;
		case VT_STREAMED_OBJECT: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {streamed object} */")); break;
		case VT_STORED_OBJECT: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {stored object} */")); break;
		case VT_BLOB_OBJECT: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {blob object} */")); break;
		case VT_CF: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {cf} */")); break;
		case VT_CLSID: StrCat(pszBuf, cchMaxBuf, _T("/* NOT IMPLEMENTED: {CLSID} */")); break;
		case VT_CARRAY:
		{
			GetTypeName(pszBuf, cchMaxBuf, tdesc.lpadesc->tdescElem, NULL, pITypeInfo);
			StrCat(pszBuf, cchMaxBuf, _T(" %ls"), pwszName); pwszName = NULL;
			for (int i = 0; i < tdesc.lpadesc->cDims; i++)
				StrCat(pszBuf, cchMaxBuf, _T("[%u]"), tdesc.lpadesc->rgbounds[i].cElements);
			break;
		}
		case VT_PTR:
		{
			GetTypeName(pszBuf, cchMaxBuf, *tdesc.lptdesc, NULL, pITypeInfo);
			StrCat(pszBuf, cchMaxBuf, _T("*"));
			break;
		}
		case VT_USERDEFINED:
		{
			if (ITypeInfo_GetRefTypeInfo(pITypeInfo, tdesc.hreftype, &pITypeInfo) == S_OK)
			{
				TYPEATTR *pTypeAttr;

				if (ITypeInfo_GetTypeAttr(pITypeInfo, &pTypeAttr) == S_OK)
				{
					BSTR pwszTypeName;

					if (ITypeInfo_GetDocumentation(pITypeInfo, MEMBERID_NIL, &pwszTypeName, 0, 0, 0) == S_OK)
					{
						if (pTypeAttr->typekind == TKIND_ENUM)
							StrCat(pszBuf, cchMaxBuf, _T("enum %ls"), pwszTypeName);
						else if (pTypeAttr->typekind == TKIND_RECORD)
							StrCat(pszBuf, cchMaxBuf, _T("struct %ls"), pwszTypeName);
						else if (pTypeAttr->typekind == TKIND_UNION)
							StrCat(pszBuf, cchMaxBuf, _T("union %ls"), pwszTypeName);
						else
							StrCat(pszBuf, cchMaxBuf, _T("%ls"), pwszTypeName);

						SysFreeString(pwszTypeName);
					}

					ITypeInfo_ReleaseTypeAttr(pITypeInfo, pTypeAttr);
				}

				ITypeInfo_Release(pITypeInfo);
			}
			break;
		}
	}

	if (pwszName != NULL)
		StrCat(pszBuf, cchMaxBuf, _T(" %ls"), pwszName);

	if (tdesc.vt & VT_ARRAY)
		StrCat(pszBuf, cchMaxBuf, _T(")"));
}

/****************************************************************************
 *																			*
 * Function: StrCat 														*
 *																			*
 * Purpose : Append text to a string (with basic size check).				*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void StrCat(PTSTR pszBuf, int cchMaxBuf, PCTSTR pcszFmt, ...)
{
	va_list args;

	va_start(args, pcszFmt);
	_vsntprintf(pszBuf + _tcslen(pszBuf), cchMaxBuf - _tcslen(pszBuf), pcszFmt, args);
	va_end(args);
}

/****************************************************************************
 *																			*
 * Function: BufCat 														*
 *																			*
 * Purpose : Append text to a dynamic buffer.								*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static void BufCat(POUTPUT pOut, PCTSTR pcszFmt, ...)
{
	static TCHAR achBuf[MAXLINE];
	va_list args;

	/* Format the text to append */
	va_start(args, pcszFmt);
	_vsntprintf(achBuf, NELEMS(achBuf), pcszFmt, args);
	va_end(args);

	/* Increase the output buffer when needed */
	if (pOut->cchBuf + _tcslen(achBuf) + 1 > pOut->cchMaxBuf)
	{
		PTSTR pchBuf = realloc(pOut->pchBuf, (pOut->cchMaxBuf + BUFSTEP) * sizeof(TCHAR));
		if (!pchBuf) return;  /* Just in case we are very unlucky... */
		pOut->pchBuf = pchBuf;
		pOut->cchMaxBuf += BUFSTEP;
	}

	/* Append the formatted text to the output buffer */
	_tcscpy(pOut->pchBuf + pOut->cchBuf, achBuf);
	pOut->cchBuf += _tcslen(achBuf);
}

/****************************************************************************
 *																			*
 * Function: AllocName														*
 *																			*
 * Purpose : Add a new (tag) name to the linked list.						*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static PNAME AllocName(const WCHAR *pcwszName)
{
	PNAME pName;

	/* Already seen this name?! */
	if ((pName = LookupName(pcwszName)) != NULL)
		return pName;

	/* Allocate a new node */
	pName = malloc(sizeof(*pName));
	if (!pName) return NULL;

	pName->pNext = NULL;
	pName->pwszName = _wcsdup(pcwszName);

	if (!g_pNameHead)
	{
		/* This is the first node. Start the list */
		g_pNameHead = pName;
	}
	else
	{
		PNAME pNameT;

		/* Find the end of the list and append the new node */
		for (pNameT = g_pNameHead; pNameT->pNext; pNameT = pNameT->pNext)
				;

		pNameT->pNext = pName;
	}

	return pName;
}

/****************************************************************************
 *																			*
 * Function: LookupName 													*
 *																			*
 * Purpose : Search for a (tag) name in the linked list.					*
 *																			*
 * History : Date		Reason												*
 *			 00/00/00	Created												*
 *																			*
 ****************************************************************************/

static PNAME LookupName(const WCHAR *pcwszName)
{
	PNAME pName;

	/* Search for the name in the linked list */
	for (pName = g_pNameHead; pName != NULL; pName = pName->pNext)
		if (wcscmp(pName->pwszName, pcwszName) == 0) return pName;

	/* Bah. Not found */
	return NULL;
}
